/*
 *  linux/drivers/net/ehea/ehea_hw.h
 *
 *  eHEA ethernet device driver for IBM eServer System p
 *
 *  (C) Copyright IBM Corp. 2006
 *
 *  Authors:
 *       Christoph Raisch <raisch@de.ibm.com>
 *       Jan-Bernd Themann <themann@de.ibm.com>
 *       Heiko-Joerg Schick <schickhj@de.ibm.com>
 *       Thomas Klein <tklein@de.ibm.com>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __EHEA_HW_H__
#define __EHEA_HW_H__

#define QPX_SQA_VALUE   EHEA_BMASK_IBM(48,63)
#define QPX_RQ1A_VALUE  EHEA_BMASK_IBM(48,63)
#define QPX_RQ2A_VALUE  EHEA_BMASK_IBM(48,63)
#define QPX_RQ3A_VALUE  EHEA_BMASK_IBM(48,63)

#define QPTEMM_OFFSET(x) offsetof(struct ehea_qptemm, x)

struct ehea_qptemm {
	u64 qpx_hcr;
	u64 qpx_c;
	u64 qpx_herr;
	u64 qpx_aer;
	u64 qpx_sqa;
	u64 qpx_sqc;
	u64 qpx_rq1a;
	u64 qpx_rq1c;
	u64 qpx_st;
	u64 qpx_aerr;
	u64 qpx_tenure;
	u64 qpx_reserved1[(0x098 - 0x058) / 8];
	u64 qpx_portp;
	u64 qpx_reserved2[(0x100 - 0x0A0) / 8];
	u64 qpx_t;
	u64 qpx_sqhp;
	u64 qpx_sqptp;
	u64 qpx_reserved3[(0x140 - 0x118) / 8];
	u64 qpx_sqwsize;
	u64 qpx_reserved4[(0x170 - 0x148) / 8];
	u64 qpx_sqsize;
	u64 qpx_reserved5[(0x1B0 - 0x178) / 8];
	u64 qpx_sigt;
	u64 qpx_wqecnt;
	u64 qpx_rq1hp;
	u64 qpx_rq1ptp;
	u64 qpx_rq1size;
	u64 qpx_reserved6[(0x220 - 0x1D8) / 8];
	u64 qpx_rq1wsize;
	u64 qpx_reserved7[(0x240 - 0x228) / 8];
	u64 qpx_pd;
	u64 qpx_scqn;
	u64 qpx_rcqn;
	u64 qpx_aeqn;
	u64 reserved49;
	u64 qpx_ram;
	u64 qpx_reserved8[(0x300 - 0x270) / 8];
	u64 qpx_rq2a;
	u64 qpx_rq2c;
	u64 qpx_rq2hp;
	u64 qpx_rq2ptp;
	u64 qpx_rq2size;
	u64 qpx_rq2wsize;
	u64 qpx_rq2th;
	u64 qpx_rq3a;
	u64 qpx_rq3c;
	u64 qpx_rq3hp;
	u64 qpx_rq3ptp;
	u64 qpx_rq3size;
	u64 qpx_rq3wsize;
	u64 qpx_rq3th;
	u64 qpx_lpn;
	u64 qpx_reserved9[(0x400 - 0x378) / 8];
	u64 reserved_ext[(0x500 - 0x400) / 8];
	u64 reserved2[(0x1000 - 0x500) / 8];
};

#define MRx_HCR_LPARID_VALID EHEA_BMASK_IBM(0, 0)

#define MRMWMM_OFFSET(x) offsetof(struct ehea_mrmwmm, x)

struct ehea_mrmwmm {
	u64 mrx_hcr;
	u64 mrx_c;
	u64 mrx_herr;
	u64 mrx_aer;
	u64 mrx_pp;
	u64 reserved1;
	u64 reserved2;
	u64 reserved3;
	u64 reserved4[(0x200 - 0x40) / 8];
	u64 mrx_ctl[64];
};

#define QPEDMM_OFFSET(x) offsetof(struct ehea_qpedmm, x)

struct ehea_qpedmm {

	u64 reserved0[(0x400) / 8];
	u64 qpedx_phh;
	u64 qpedx_ppsgp;
	u64 qpedx_ppsgu;
	u64 qpedx_ppdgp;
	u64 qpedx_ppdgu;
	u64 qpedx_aph;
	u64 qpedx_apsgp;
	u64 qpedx_apsgu;
	u64 qpedx_apdgp;
	u64 qpedx_apdgu;
	u64 qpedx_apav;
	u64 qpedx_apsav;
	u64 qpedx_hcr;
	u64 reserved1[4];
	u64 qpedx_rrl0;
	u64 qpedx_rrrkey0;
	u64 qpedx_rrva0;
	u64 reserved2;
	u64 qpedx_rrl1;
	u64 qpedx_rrrkey1;
	u64 qpedx_rrva1;
	u64 reserved3;
	u64 qpedx_rrl2;
	u64 qpedx_rrrkey2;
	u64 qpedx_rrva2;
	u64 reserved4;
	u64 qpedx_rrl3;
	u64 qpedx_rrrkey3;
	u64 qpedx_rrva3;
};

#define CQX_FECADDER EHEA_BMASK_IBM(32, 63)
#define CQX_FEC_CQE_CNT EHEA_BMASK_IBM(32, 63)
#define CQX_N1_GENERATE_COMP_EVENT EHEA_BMASK_IBM(0, 0)
#define CQX_EP_EVENT_PENDING EHEA_BMASK_IBM(0, 0)

#define CQTEMM_OFFSET(x) offsetof(struct ehea_cqtemm, x)

struct ehea_cqtemm {
	u64 cqx_hcr;
	u64 cqx_c;
	u64 cqx_herr;
	u64 cqx_aer;
	u64 cqx_ptp;
	u64 cqx_tp;
	u64 cqx_fec;
	u64 cqx_feca;
	u64 cqx_ep;
	u64 cqx_eq;
	u64 reserved1;
	u64 cqx_n0;
	u64 cqx_n1;
	u64 reserved2[(0x1000 - 0x60) / 8];
};

#define EQTEMM_OFFSET(x) offsetof(struct ehea_eqtemm, x)

struct ehea_eqtemm {
	u64 EQx_HCR;
	u64 EQx_C;
	u64 EQx_HERR;
	u64 EQx_AER;
	u64 EQx_PTP;
	u64 EQx_TP;
	u64 EQx_SSBA;
	u64 EQx_PSBA;
	u64 EQx_CEC;
	u64 EQx_MEQL;
	u64 EQx_XISBI;
	u64 EQx_XISC;
	u64 EQx_IT;
};

static inline u64 hipz_galpa_load(struct h_galpa galpa, u32 offset)
{
	u64 addr = galpa.fw_handle + offset;
	u64 out;
	EDEB_EN(7, "addr=%lx offset=%x ", addr, offset);
	out = *(volatile u64 *)addr;
	EDEB_EX(7, "addr=%lx value=%lx", addr, out);
	return out;
};

static inline void hipz_galpa_store(struct h_galpa galpa, u32 offset, u64 value)
{
	u64 addr = galpa.fw_handle + offset;
	EDEB_EN(7, "addr=%lx offset=%x value=%lx", addr, offset,
		value);
	*(u64 *) addr = value;
	hipz_galpa_load(galpa, offset);	/* synchronize explicitly to ehea */
	EDEB_EX(7, "");
};

static inline void hipz_galpa_store_acc(struct h_galpa galpa, u32 offset,
					u64 value)
{
	u64 addr = galpa.fw_handle + offset;
	EDEB(7, "Accelerated store: addr=%lx offset=%x value=%lx",
	     addr, offset, value);
	*(u64 *) addr = value;
};

#define hipz_galpa_store_eq(gal, offset, value)\
        hipz_galpa_store(gal, EQTEMM_OFFSET(offset), value)
#define hipz_galpa_load_eq(gal, offset)\
        hipz_galpa_load(gal, EQTEMM_OFFSET(offset))

#define hipz_galpa_store_cq(gal, offset, value)\
        hipz_galpa_store(gal, CQTEMM_OFFSET(offset), value)
#define hipz_galpa_load_cq(gal, offset)\
        hipz_galpa_load(gal, CQTEMM_OFFSET(offset))

#define hipz_galpa_store_qp(gal, offset, value)\
        hipz_galpa_store(gal, QPTEMM_OFFSET(offset), value)
#define hipz_galpa_load_qp(gal, offset)\
        hipz_galpa_load(gal, QPTEMM_OFFSET(offset))

#define hipz_galpa_store_qped(gal, offset, value)\
        hipz_galpa_store(gal, QPEDMM_OFFSET(offset), value)
#define hipz_galpa_load_qped(gal, offset)\
        hipz_galpa_load(gal, QPEDMM_OFFSET(offset))

#define hipz_galpa_store_mrmw(gal, offset, value)\
        hipz_galpa_store(gal, MRMWMM_OFFSET(offset), value)
#define hipz_galpa_load_mrmw(gal, offset)\
        hipz_galpa_load(gal, MRMWMM_OFFSET(offset))

#define hipz_galpa_store_base(gal, offset, value)\
        hipz_galpa_store(gal, HCAGR_OFFSET(offset), value)
#define hipz_galpa_load_base(gal, offset)\
        hipz_galpa_load(gal, HCAGR_OFFSET(offset))

static inline void ehea_update_sqa(struct ehea_qp *qp, u16 nr_wqes)
{
	struct h_galpa gal = qp->galpas.kernel;
	EDEB_EN(7, "qp=%p, nr_wqes=%d", qp, nr_wqes);

	hipz_galpa_store_acc(gal, QPTEMM_OFFSET(qpx_sqa),
			     EHEA_BMASK_SET(QPX_SQA_VALUE, nr_wqes));
	EDEB_EX(7, "qpx_sqa = %i", nr_wqes);
}

static inline void ehea_update_rq3a(struct ehea_qp *qp, u16 nr_wqes)
{
	struct h_galpa gal = qp->galpas.kernel;
	EDEB_EN(7, "ehea_qp=%p, nr_wqes=%d", qp, nr_wqes);
	hipz_galpa_store_acc(gal, QPTEMM_OFFSET(qpx_rq3a),
			     EHEA_BMASK_SET(QPX_RQ1A_VALUE, nr_wqes));
	EDEB_EX(7, "QPx_RQA = %i", nr_wqes);
}

static inline void ehea_update_rq2a(struct ehea_qp *qp, u16 nr_wqes)
{
	struct h_galpa gal = qp->galpas.kernel;
	EDEB_EN(7, "ehea_qp=%p, nr_wqes=%d", qp, nr_wqes);
	hipz_galpa_store_acc(gal, QPTEMM_OFFSET(qpx_rq2a),
			     EHEA_BMASK_SET(QPX_RQ1A_VALUE, nr_wqes));
	EDEB_EX(7, "QPx_RQA = %i", nr_wqes);
}

static inline void ehea_update_rq1a(struct ehea_qp *qp, u16 nr_wqes)
{
	struct h_galpa gal = qp->galpas.kernel;
	EDEB_EN(7, "ehea_qp=%p, nr_wqes=%d", qp, nr_wqes);
	hipz_galpa_store_acc(gal, QPTEMM_OFFSET(qpx_rq1a),
			     EHEA_BMASK_SET(QPX_RQ1A_VALUE, nr_wqes));
	EDEB_EX(7, "QPx_RQA = %i", nr_wqes);
}

static inline void ehea_update_feca(struct ehea_cq *cq, u32 nr_cqes)
{
	struct h_galpa gal = cq->galpas.kernel;
	EDEB_EN(7, "");
	hipz_galpa_store_acc(gal, CQTEMM_OFFSET(cqx_feca),
			     EHEA_BMASK_SET(CQX_FECADDER, nr_cqes));
	EDEB_EX(7, "cqx_feca = %i", nr_cqes);
}

static inline void ehea_reset_cq_n1(struct ehea_cq *cq)
{
	struct h_galpa gal = cq->galpas.kernel;
	EDEB_EN(7, "");
	hipz_galpa_store_cq(gal,
			    cqx_n1,
			    EHEA_BMASK_SET(CQX_N1_GENERATE_COMP_EVENT, 1));
	EDEB_EX(7, "");
}

static inline void ehea_reset_cq_ep(struct ehea_cq *my_cq)
{
	struct h_galpa gal = my_cq->galpas.kernel;
	EDEB_EN(7, "");
	hipz_galpa_store_acc(gal,
			     CQTEMM_OFFSET(cqx_ep),
			     EHEA_BMASK_SET(CQX_EP_EVENT_PENDING, 0));
	EDEB_EX(7, "");
}


#endif	/* __EHEA_HW_H__ */
